.ui-button {
  display: flex;
  align-items: center;
  color: currentcolor;
  cursor: pointer;

  &.icon {
    appearance: none;
    padding: 0;

    background: transparent;
    border: 0px solid transparent;
  }
}
