/* Common styles for both error and warn */
.container
  padding: 10px

.header
  font-weight: bold
  display: flex
  justify-content: space-between

.details
  white-space: pre-wrap
  word-wrap: break-word
  text-indent: 0
  margin-bottom: 16px;
  margin-top: 32px;
